from __future__ import annotations

from collections.abc import Mapping, Sequence
from typing import Any, Literal, Union
from uuid import UUID

from pydantic import AliasChoices
from pydantic import BaseModel, ConfigDict, Field
from pydantic import field_validator
from pydantic import model_validator


class RuntimeRerankControl(BaseModel):
    enabled: bool | None = None


class RuntimeQueryExpansionControl(BaseModel):
    enabled: bool | None = None


class RuntimeSubquestionHitlControl(BaseModel):
    enabled: bool = False


class RuntimeHitlControl(BaseModel):
    enabled: bool = False
    subquestions: RuntimeSubquestionHitlControl | None = None

    @model_validator(mode="after")
    def normalize_enabled(self) -> "RuntimeHitlControl":
        if self.subquestions is not None and self.subquestions.enabled:
            self.enabled = True
        return self


class RuntimeAgentRunControls(BaseModel):
    rerank: RuntimeRerankControl | None = None
    query_expansion: RuntimeQueryExpansionControl | None = None
    hitl: RuntimeHitlControl | None = None


class RuntimeAgentRunRuntimeConfig(BaseModel):
    rerank: RuntimeRerankControl | None = None
    query_expansion: RuntimeQueryExpansionControl | None = None


class RuntimeCustomPrompts(BaseModel):
    model_config = ConfigDict(extra="ignore")

    subanswer: str | None = None
    synthesis: str | None = None


class RuntimeAgentRunRequest(BaseModel):
    query: str = Field(min_length=1)
    thread_id: str | None = None
    checkpoint_db_url: str | None = None
    controls: RuntimeAgentRunControls | None = None
    runtime_config: RuntimeAgentRunRuntimeConfig | None = None
    custom_prompts: RuntimeCustomPrompts | None = Field(
        default=None,
        validation_alias=AliasChoices("custom_prompts", "custom-prompts"),
    )

    @field_validator("thread_id")
    @classmethod
    def validate_thread_id(cls, value: str | None) -> str | None:
        if value is None:
            return None
        normalized = value.strip()
        if not normalized:
            return None
        return str(UUID(normalized))


class SubQuestionAnswer(BaseModel):
    sub_question: str
    sub_answer: str
    tool_call_input: str = ""
    expanded_query: str = ""
    sub_agent_response: str = ""
    answerable: bool = False
    verification_reason: str = ""


class RuntimeAgentRunResponse(BaseModel):
    main_question: str = ""
    thread_id: str | None = None
    sub_items: list[tuple[str, str]] = Field(default_factory=list)
    output: str
    final_citations: list["CitationSourceRow"] = Field(default_factory=list)


class HitlResumeDecision(BaseModel):
    item_id: str = Field(min_length=1)
    action: Literal["approve", "edit", "reject"]
    replacement_text: str | None = None

    @model_validator(mode="after")
    def validate_replacement_text(self) -> "HitlResumeDecision":
        if self.action == "edit":
            if self.replacement_text is None or not self.replacement_text.strip():
                raise ValueError("replacement_text is required when action='edit'.")
            self.replacement_text = self.replacement_text.strip()
            return self
        if self.replacement_text is not None:
            self.replacement_text = self.replacement_text.strip() or None
        return self


class HitlReviewItem(BaseModel):
    item_id: str = Field(min_length=1)
    text: str
    index: int = Field(ge=0)

    def approve(self) -> HitlResumeDecision:
        return HitlResumeDecision(item_id=self.item_id, action="approve")

    def edit(self, replacement_text: str) -> HitlResumeDecision:
        return HitlResumeDecision(item_id=self.item_id, action="edit", replacement_text=replacement_text)

    def reject(self) -> HitlResumeDecision:
        return HitlResumeDecision(item_id=self.item_id, action="reject")


class HitlReview(BaseModel):
    kind: Literal["subquestion_review"]
    stage: str = ""
    checkpoint_id: str = Field(min_length=1)
    items: list[HitlReviewItem] = Field(default_factory=list)

    @classmethod
    def from_interrupt_payload(cls, payload: Any) -> "HitlReview":
        if not isinstance(payload, Mapping):
            raise ValueError("HITL interrupt payload must be a mapping.")
        kind = str(payload.get("kind") or "").strip()
        checkpoint_id = str(payload.get("checkpoint_id") or "").strip()
        if not kind:
            raise ValueError("HITL interrupt payload is missing kind.")
        if kind != "subquestion_review":
            raise ValueError(f"Unsupported HITL review kind '{kind}'.")
        if not checkpoint_id:
            raise ValueError("HITL interrupt payload is missing checkpoint_id.")
        return cls(
            kind=kind,
            stage=str(payload.get("stage") or ""),
            checkpoint_id=checkpoint_id,
            items=cls._normalize_items(kind=kind, payload=payload),
        )

    @staticmethod
    def _normalize_items(*, kind: str, payload: Mapping[str, Any]) -> list[HitlReviewItem]:
        raw_items_key = "subquestions"
        raw_items = payload.get(raw_items_key)
        if not isinstance(raw_items, Sequence) or isinstance(raw_items, (str, bytes, bytearray)):
            return []
        item_id_key = "subquestion_id"
        text_key = "sub_question"
        items: list[HitlReviewItem] = []
        for index, raw_item in enumerate(raw_items):
            if not isinstance(raw_item, Mapping):
                continue
            raw_index = raw_item.get("index")
            items.append(
                HitlReviewItem(
                    item_id=str(raw_item.get(item_id_key) or f"item-{index + 1}").strip(),
                    text=str(raw_item.get(text_key) or "").strip(),
                    index=raw_index if isinstance(raw_index, int) and raw_index >= 0 else index,
                )
            )
        return items

    def approve_all(self) -> "HitlResumeRequest":
        return HitlResumeRequest(
            checkpoint_id=self.checkpoint_id,
            review_kind=self.kind,
            decisions=[item.approve() for item in self.items],
        )

    def with_decisions(self, *decisions: HitlResumeDecision) -> "HitlResumeRequest":
        return HitlResumeRequest(
            checkpoint_id=self.checkpoint_id,
            review_kind=self.kind,
            decisions=list(decisions),
        )


class HitlResumeRequest(BaseModel):
    checkpoint_id: str = Field(min_length=1)
    review_kind: Literal["subquestion_review"]
    decisions: list[HitlResumeDecision] = Field(min_length=1)

    @classmethod
    def approve(cls, review: HitlReview, item_id: str) -> "HitlResumeRequest":
        return review.with_decisions(HitlResumeDecision(item_id=item_id, action="approve"))

    @classmethod
    def edit(cls, review: HitlReview, item_id: str, replacement_text: str) -> "HitlResumeRequest":
        return review.with_decisions(
            HitlResumeDecision(item_id=item_id, action="edit", replacement_text=replacement_text)
        )

    @classmethod
    def reject(cls, review: HitlReview, item_id: str) -> "HitlResumeRequest":
        return review.with_decisions(HitlResumeDecision(item_id=item_id, action="reject"))


class RuntimeAgentRunResult(BaseModel):
    status: Literal["completed", "paused"]
    checkpoint_id: str | None = None
    review: HitlReview | None = Field(
        default=None,
        validation_alias=AliasChoices("review", "interrupt_payload"),
    )
    response: RuntimeAgentRunResponse | None = None

    @field_validator("review", mode="before")
    @classmethod
    def validate_review(cls, value: Any) -> HitlReview | Any | None:
        if value is None or isinstance(value, HitlReview):
            return value
        if isinstance(value, Mapping):
            return HitlReview.from_interrupt_payload(value)
        return value


class AgentRunStageMetadata(BaseModel):
    stage: str
    status: str
    sub_question: str = ""
    lane_index: int = 0
    lane_total: int = 0
    emitted_at: float | None = None


class RuntimeAgentRunAsyncStartResponse(BaseModel):
    job_id: str
    run_id: str
    thread_id: str = ""
    status: str


class RuntimeAgentRunAsyncStatusResponse(BaseModel):
    job_id: str
    run_id: str = ""
    thread_id: str = ""
    status: str
    message: str = ""
    stage: str = ""
    stages: list[AgentRunStageMetadata] = Field(default_factory=list)
    decomposition_sub_questions: list[str] = Field(default_factory=list)
    sub_question_artifacts: list["SubQuestionArtifacts"] = Field(default_factory=list)
    sub_items: list[tuple[str, str]] = Field(default_factory=list)
    output: str = ""
    result: RuntimeAgentRunResponse | None = None
    error: str | None = None
    cancel_requested: bool = False
    review: HitlReview | None = Field(
        default=None,
        validation_alias=AliasChoices("review", "interrupt_payload"),
    )
    checkpoint_id: str | None = None
    started_at: float | None = None
    finished_at: float | None = None
    elapsed_ms: int | None = None

    @field_validator("review", mode="before")
    @classmethod
    def validate_review(cls, value: Any) -> HitlReview | Any | None:
        if value is None or isinstance(value, HitlReview):
            return value
        if isinstance(value, Mapping):
            return HitlReview.from_interrupt_payload(value)
        return value

class RuntimeAgentRunAsyncCancelResponse(BaseModel):
    status: Literal["success"]
    message: str


class RuntimeSubquestionDecision(BaseModel):
    subquestion_id: str = Field(min_length=1)
    action: Literal["approve", "edit", "deny", "skip"]
    edited_text: str | None = None

    @model_validator(mode="after")
    def validate_edit_payload(self) -> "RuntimeSubquestionDecision":
        if self.action == "edit":
            if self.edited_text is None or not self.edited_text.strip():
                raise ValueError("edited_text is required when action='edit'.")
        elif self.edited_text is not None:
            self.edited_text = self.edited_text.strip() or None
        return self


class RuntimeSubquestionResumeEnvelope(BaseModel):
    checkpoint_id: str = Field(min_length=1)
    decisions: list[RuntimeSubquestionDecision] = Field(min_length=1)


class RuntimeAgentRunResumeRequest(BaseModel):
    resume: Union[
        bool,
        HitlResumeRequest,
        dict[str, Any],
        RuntimeSubquestionResumeEnvelope,
    ] = True

    @field_validator("resume", mode="before")
    @classmethod
    def validate_resume(cls, value: Any) -> Any:
        if isinstance(value, bool):
            return value
        if isinstance(value, HitlResumeRequest):
            return value
        if not isinstance(value, dict):
            raise ValueError(
                "resume must be a boolean, SDK HITL resume request, legacy object payload, or typed decision envelope."
            )
        if "review_kind" in value and "checkpoint_id" in value and "decisions" in value:
            return HitlResumeRequest.model_validate(value)
        if "checkpoint_id" in value or "decisions" in value:
            return RuntimeSubquestionResumeEnvelope.model_validate(value)
        return value


class RuntimeAgentInfo(BaseModel):
    agent_name: str


class GraphRunMetadata(BaseModel):
    run_id: str
    thread_id: str = ""
    trace_id: str = ""
    correlation_id: str = ""


class CitationSourceRow(BaseModel):
    citation_index: int = Field(ge=1)
    rank: int = Field(ge=1)
    title: str = ""
    source: str = ""
    content: str = ""
    document_id: str = ""
    score: float | None = None


class SubQuestionArtifacts(BaseModel):
    sub_question: str
    expanded_queries: list[str] = Field(default_factory=list)
    retrieved_docs: list[CitationSourceRow] = Field(default_factory=list)
    retrieval_provenance: list[dict[str, Any]] = Field(default_factory=list)
    reranked_docs: list[CitationSourceRow] = Field(default_factory=list)
    sub_answer: str = ""
    citation_rows_by_index: dict[int, CitationSourceRow] = Field(default_factory=dict)


class AgentGraphState(BaseModel):
    main_question: str
    decomposition_sub_questions: list[str] = Field(default_factory=list)
    sub_question_artifacts: list[SubQuestionArtifacts] = Field(default_factory=list)
    final_answer: str = ""
    citation_rows_by_index: dict[int, CitationSourceRow] = Field(default_factory=dict)
    run_metadata: GraphRunMetadata
    sub_qa: list[SubQuestionAnswer] = Field(default_factory=list)
    output: str = ""
    stage_snapshots: list["GraphStageSnapshot"] = Field(default_factory=list)


class DecomposeNodeInput(BaseModel):
    main_question: str
    run_metadata: GraphRunMetadata
    initial_search_context: list[dict[str, Any]] = Field(default_factory=list)


class DecomposeNodeOutput(BaseModel):
    decomposition_sub_questions: list[str] = Field(default_factory=list)


class ExpandNodeInput(BaseModel):
    main_question: str
    sub_question: str
    run_metadata: GraphRunMetadata


class ExpandNodeOutput(BaseModel):
    expanded_queries: list[str] = Field(default_factory=list)


class SearchNodeInput(BaseModel):
    sub_question: str
    expanded_queries: list[str] = Field(default_factory=list)
    run_metadata: GraphRunMetadata


class SearchNodeOutput(BaseModel):
    retrieved_docs: list[CitationSourceRow] = Field(default_factory=list)
    retrieval_provenance: list[dict[str, Any]] = Field(default_factory=list)
    citation_rows_by_index: dict[int, CitationSourceRow] = Field(default_factory=dict)


class RerankNodeInput(BaseModel):
    sub_question: str
    expanded_query: str = ""
    retrieved_docs: list[CitationSourceRow] = Field(default_factory=list)
    run_metadata: GraphRunMetadata


class RerankNodeOutput(BaseModel):
    reranked_docs: list[CitationSourceRow] = Field(default_factory=list)
    citation_rows_by_index: dict[int, CitationSourceRow] = Field(default_factory=dict)


class AnswerSubquestionNodeInput(BaseModel):
    sub_question: str
    reranked_docs: list[CitationSourceRow] = Field(default_factory=list)
    citation_rows_by_index: dict[int, CitationSourceRow] = Field(default_factory=dict)
    run_metadata: GraphRunMetadata


class AnswerSubquestionNodeOutput(BaseModel):
    sub_answer: str = ""
    citation_indices_used: list[int] = Field(default_factory=list)
    answerable: bool = False
    verification_reason: str = ""
    citation_rows_by_index: dict[int, CitationSourceRow] = Field(default_factory=dict)


class SynthesizeFinalNodeInput(BaseModel):
    main_question: str
    sub_qa: list[SubQuestionAnswer] = Field(default_factory=list)
    sub_question_artifacts: list[SubQuestionArtifacts] = Field(default_factory=list)
    run_metadata: GraphRunMetadata


class SynthesizeFinalNodeOutput(BaseModel):
    final_answer: str = ""


class GraphStageSnapshot(BaseModel):
    stage: str
    status: str = "completed"
    sub_question: str = ""
    lane_index: int = 0
    lane_total: int = 0
    decomposition_sub_questions: list[str] = Field(default_factory=list)
    sub_qa: list[SubQuestionAnswer] = Field(default_factory=list)
    sub_question_artifacts: list[SubQuestionArtifacts] = Field(default_factory=list)
    output: str = ""



from __future__ import annotations

import logging
from typing import Any

from langchain_community.vectorstores.pgvector import PGVector, _get_embedding_collection_store
from langchain_core.documents import Document
from langchain_core.embeddings import Embeddings
from sqlalchemy import create_engine, inspect
from sqlalchemy.orm import Session

logger = logging.getLogger(__name__)

CITATION_TITLE_METADATA_KEY = "citation_title"
CITATION_SOURCE_METADATA_KEY = "citation_source"
CITATION_DOCUMENT_ID_METADATA_KEY = "document_id"


def _collection_exists(connection: str, collection_name: str, use_jsonb: bool = True) -> bool:
    """Check whether the target PGVector collection already exists."""
    _, collection_store = _get_embedding_collection_store(None, use_jsonb=use_jsonb)
    engine = create_engine(connection, future=True)
    try:
        inspector = inspect(engine)
        if not inspector.has_table(collection_store.__tablename__):
            return False
        with Session(engine) as session:
            return collection_store.get_by_name(session, collection_name) is not None
    finally:
        engine.dispose()


def get_vector_store(connection: str, collection_name: str, embeddings: Embeddings) -> PGVector:
    """Create or fetch a single PGVector collection for app use."""
    existed_before = _collection_exists(connection=connection, collection_name=collection_name, use_jsonb=True)
    vector_store = PGVector(
        connection_string=connection,
        embedding_function=embeddings,
        collection_name=collection_name,
        use_jsonb=True,
    )
    logger.info(
        "PGVector collection ready: collection_name='%s' state=%s",
        collection_name,
        "existing" if existed_before else "created",
    )
    return vector_store


def _normalize_document_metadata(document: Document) -> Document:
    """Keep only required citation metadata fields for retrieval/filtering."""
    metadata = document.metadata or {}
    topic = str(
        metadata.get(CITATION_TITLE_METADATA_KEY)
        or metadata.get("title")
        or metadata.get("wiki_page")
        or ""
    ).strip()
    wiki_url = str(
        metadata.get(CITATION_SOURCE_METADATA_KEY)
        or metadata.get("source")
        or metadata.get("wiki_url")
        or ""
    ).strip()
    document_id = str(metadata.get(CITATION_DOCUMENT_ID_METADATA_KEY) or document.id or "").strip()
    # Preserve explicit citation keys and legacy wiki keys for non-citation consumers.
    slim_metadata = {
        CITATION_TITLE_METADATA_KEY: topic,
        CITATION_SOURCE_METADATA_KEY: wiki_url,
        CITATION_DOCUMENT_ID_METADATA_KEY: document_id,
        "topic": topic,
        "wiki_url": wiki_url,
        "wiki_page": topic,
    }
    return Document(page_content=document.page_content, metadata=slim_metadata, id=document.id)


def add_documents_to_store(vector_store: PGVector, documents: list[Document]) -> list[str]:
    """Add documents to PGVector and return generated IDs."""
    if not documents:
        logger.info("No documents provided; skipping PGVector add.")
        return []

    normalized_documents = [_normalize_document_metadata(document) for document in documents]
    ids = vector_store.add_documents(normalized_documents)
    logger.info(
        "Added %s documents to PGVector collection='%s'",
        len(ids),
        vector_store.collection_name,
    )
    return ids


def search_documents_for_context(
    vector_store: Any,
    query: str,
    *,
    k: int,
    score_threshold: float | None = None,
) -> list[Document]:
    """Run one retrieval pass for context gathering before decomposition."""
    safe_k = max(1, k)
    if score_threshold is not None and hasattr(vector_store, "similarity_search_with_relevance_scores"):
        docs_with_scores = vector_store.similarity_search_with_relevance_scores(
            query,
            k=safe_k,
            score_threshold=score_threshold,
        )
        documents: list[Document] = []
        for doc, score in docs_with_scores:
            metadata = dict(doc.metadata or {})
            metadata["score"] = score
            documents.append(Document(page_content=doc.page_content, metadata=metadata, id=doc.id))
        logger.info(
            "Context search complete query=%r k=%s score_threshold=%s results=%s mode=with_scores",
            query,
            safe_k,
            score_threshold,
            len(documents),
        )
        return documents

    documents = vector_store.similarity_search(query, k=safe_k)
    logger.info(
        "Context search complete query=%r k=%s score_threshold=%s results=%s mode=similarity_search",
        query,
        safe_k,
        score_threshold,
        len(documents),
    )
    return documents


def search_documents_for_queries(
    vector_store: Any,
    queries: list[str],
    *,
    k: int,
    score_threshold: float | None = None,
) -> dict[str, list[Document]]:
    """Run retrieval for each query and return results keyed by query text."""
    results: dict[str, list[Document]] = {}
    safe_queries = [query.strip() for query in queries if isinstance(query, str) and query.strip()]
    logger.info(
        "Multi-query context search start query_count=%s k=%s score_threshold=%s",
        len(safe_queries),
        max(1, k),
        score_threshold,
    )
    for query in safe_queries:
        documents = search_documents_for_context(
            vector_store=vector_store,
            query=query,
            k=k,
            score_threshold=score_threshold,
        )
        results[query] = documents
    logger.info(
        "Multi-query context search complete query_count=%s total_results=%s",
        len(safe_queries),
        sum(len(docs) for docs in results.values()),
    )
    return results


def build_initial_search_context(documents: list[Document]) -> list[dict[str, str | int]]:
    """Return bounded, structured context items for coordinator decomposition input."""
    context_items: list[dict[str, str | int]] = []
    for rank, document in enumerate(documents, start=1):
        metadata = document.metadata or {}
        snippet = document.page_content.strip().replace("\n", " ")
        context_items.append(
            {
                "rank": rank,
                "document_id": str(metadata.get(CITATION_DOCUMENT_ID_METADATA_KEY) or document.id or ""),
                "title": str(metadata.get(CITATION_TITLE_METADATA_KEY) or ""),
                "source": str(metadata.get(CITATION_SOURCE_METADATA_KEY) or ""),
                "snippet": snippet[:250],
            }
        )
    return context_items



from __future__ import annotations

from dataclasses import dataclass
from typing import Iterator, Literal, Mapping

from pydantic import BaseModel

from schemas import (
    AnswerSubquestionNodeInput,
    AnswerSubquestionNodeOutput,
    DecomposeNodeInput,
    DecomposeNodeOutput,
    ExpandNodeInput,
    ExpandNodeOutput,
    RerankNodeInput,
    RerankNodeOutput,
    SearchNodeInput,
    SearchNodeOutput,
    SynthesizeFinalNodeInput,
    SynthesizeFinalNodeOutput,
)

NodeName = Literal[
    "decompose",
    "expand",
    "search",
    "rerank",
    "answer_subquestion",
    "synthesize_final",
]


@dataclass(frozen=True, slots=True)
class NodeIOContract:
    node_name: NodeName
    input_schema: type[BaseModel]
    output_schema: type[BaseModel]
    implementation: str


NODE_IO_CONTRACTS: Mapping[NodeName, NodeIOContract] = {
    "decompose": NodeIOContract(
        node_name="decompose",
        input_schema=DecomposeNodeInput,
        output_schema=DecomposeNodeOutput,
        implementation="agent_search.runtime.nodes.decompose.run_decomposition_node",
    ),
    "expand": NodeIOContract(
        node_name="expand",
        input_schema=ExpandNodeInput,
        output_schema=ExpandNodeOutput,
        implementation="agent_search.runtime.nodes.expand.run_expansion_node",
    ),
    "search": NodeIOContract(
        node_name="search",
        input_schema=SearchNodeInput,
        output_schema=SearchNodeOutput,
        implementation="agent_search.runtime.nodes.search.run_search_node",
    ),
    "rerank": NodeIOContract(
        node_name="rerank",
        input_schema=RerankNodeInput,
        output_schema=RerankNodeOutput,
        implementation="agent_search.runtime.nodes.rerank.run_rerank_node",
    ),
    "answer_subquestion": NodeIOContract(
        node_name="answer_subquestion",
        input_schema=AnswerSubquestionNodeInput,
        output_schema=AnswerSubquestionNodeOutput,
        implementation="agent_search.runtime.nodes.answer.run_answer_node",
    ),
    "synthesize_final": NodeIOContract(
        node_name="synthesize_final",
        input_schema=SynthesizeFinalNodeInput,
        output_schema=SynthesizeFinalNodeOutput,
        implementation="agent_search.runtime.nodes.synthesize.run_synthesize_node",
    ),
}


def iter_node_io_contracts() -> Iterator[NodeIOContract]:
    return iter(NODE_IO_CONTRACTS.values())


def get_node_io_contract(node_name: NodeName) -> NodeIOContract:
    return NODE_IO_CONTRACTS[node_name]


__all__ = [
    "NODE_IO_CONTRACTS",
    "NodeIOContract",
    "NodeName",
    "get_node_io_contract",
    "iter_node_io_contracts",
]



from __future__ import annotations

import logging
import re
from typing import Any, Callable

from schemas import AnswerSubquestionNodeInput, AnswerSubquestionNodeOutput, CitationSourceRow
from services.document_validation_service import RetrievedDocument, format_retrieved_documents
from services.subanswer_service import generate_subanswer

logger = logging.getLogger(__name__)

_QUERY_LOG_MAX = 200
_ANSWER_SUBQUESTION_NO_SUPPORT_FALLBACK = "nothing relevant found"
_CITATION_INDEX_PATTERN = re.compile(r"\[(\d+)\]")


def _truncate_query(q: str) -> str:
    return q[: _QUERY_LOG_MAX] + "..." if len(q) > _QUERY_LOG_MAX else q


def _extract_citation_indices(answer: str) -> list[int]:
    if not answer:
        return []
    seen: set[int] = set()
    indices: list[int] = []
    for raw_index in _CITATION_INDEX_PATTERN.findall(answer):
        try:
            index = int(raw_index)
        except ValueError:
            continue
        if index <= 0 or index in seen:
            continue
        seen.add(index)
        indices.append(index)
    return indices


def _format_citation_rows_for_pipeline(rows: list[CitationSourceRow]) -> str:
    documents = [
        RetrievedDocument(
            rank=row.rank,
            title=row.title,
            source=row.source,
            content=row.content,
        )
        for row in rows
    ]
    return format_retrieved_documents(documents)


def run_answer_node(
    *,
    node_input: AnswerSubquestionNodeInput,
    callbacks: list[Any] | None = None,
    no_support_fallback: str = _ANSWER_SUBQUESTION_NO_SUPPORT_FALLBACK,
    format_citation_rows_for_pipeline_fn: Callable[[list[CitationSourceRow]], str] = _format_citation_rows_for_pipeline,
    generate_subanswer_fn: Callable[..., str] = generate_subanswer,
    extract_citation_indices_fn: Callable[[str], list[int]] = _extract_citation_indices,
    truncate_query_fn: Callable[[str], str] = _truncate_query,
) -> AnswerSubquestionNodeOutput:
    logger.info(
        "Subanswer node start sub_question=%s reranked_doc_count=%s run_id=%s trace_id=%s correlation_id=%s",
        truncate_query_fn(node_input.sub_question),
        len(node_input.reranked_docs),
        node_input.run_metadata.run_id,
        node_input.run_metadata.trace_id,
        node_input.run_metadata.correlation_id,
    )

    if not node_input.reranked_docs:
        logger.info(
            "Subanswer node fallback; no reranked docs sub_question=%s run_id=%s",
            truncate_query_fn(node_input.sub_question),
            node_input.run_metadata.run_id,
        )
        return AnswerSubquestionNodeOutput(
            sub_answer=no_support_fallback,
            citation_indices_used=[],
            answerable=False,
            verification_reason="no_reranked_documents",
            citation_rows_by_index={},
        )

    reranked_output = format_citation_rows_for_pipeline_fn(node_input.reranked_docs)
    generated_sub_answer = generate_subanswer_fn(
        sub_question=node_input.sub_question,
        reranked_retrieved_output=reranked_output,
        callbacks=callbacks,
    )

    citation_rows = dict(node_input.citation_rows_by_index)
    if not citation_rows:
        citation_rows = {row.citation_index: row for row in node_input.reranked_docs}
    citation_indices_used = extract_citation_indices_fn(generated_sub_answer)
    supports_answer = True
    verification_reason = "citation_supported"
    invalid_indices = [index for index in citation_indices_used if index not in citation_rows]
    missing_citations = supports_answer and not citation_indices_used
    missing_support_rows = supports_answer and bool(citation_indices_used) and bool(invalid_indices)

    if missing_citations:
        supports_answer = False
        verification_reason = "missing_citation_markers"
    elif missing_support_rows:
        supports_answer = False
        verification_reason = "missing_supporting_source_rows"

    if not supports_answer:
        logger.info(
            "Subanswer node fallback; unsupported answer sub_question=%s reason=%s citation_indices=%s run_id=%s",
            truncate_query_fn(node_input.sub_question),
            verification_reason,
            citation_indices_used,
            node_input.run_metadata.run_id,
        )
        return AnswerSubquestionNodeOutput(
            sub_answer=no_support_fallback,
            citation_indices_used=[],
            answerable=False,
            verification_reason=verification_reason,
            citation_rows_by_index={},
        )

    supporting_rows = {
        index: citation_rows[index].model_copy(deep=True)
        for index in citation_indices_used
        if index in citation_rows
    }
    logger.info(
        "Subanswer node complete sub_question=%s answer_len=%s citation_count=%s run_id=%s",
        truncate_query_fn(node_input.sub_question),
        len(generated_sub_answer),
        len(citation_indices_used),
        node_input.run_metadata.run_id,
    )
    return AnswerSubquestionNodeOutput(
        sub_answer=generated_sub_answer,
        citation_indices_used=citation_indices_used,
        answerable=True,
        verification_reason=verification_reason,
        citation_rows_by_index=supporting_rows,
    )


__all__ = ["run_answer_node"]



from __future__ import annotations

from typing import Any

from langgraph.types import Command


class ResumeTransitionError(ValueError):
    """Raised when a run is resumed from an invalid lifecycle state."""


_RESUMABLE_STATUSES = frozenset({"paused"})


def build_resume_command(resume: Any = True) -> Command:
    return Command(resume=resume)


def ensure_resume_allowed(status: str) -> None:
    normalized_status = status.strip().lower()
    if normalized_status in _RESUMABLE_STATUSES:
        return
    raise ResumeTransitionError(f"Run cannot be resumed from status '{status}'.")


__all__ = [
    "ResumeTransitionError",
    "build_resume_command",
    "ensure_resume_allowed",
]



from __future__ import annotations

import inspect
import logging
import threading
import uuid
from typing import Any, Iterable, Literal, Mapping

from config import LangfuseSettings, should_sample_rate

logger = logging.getLogger(__name__)

_LangfuseScope = Literal["runtime"]
_UNSET: object = object()
_cached_langfuse_client: Any | object = _UNSET
_cache_lock = threading.Lock()
TRACE_CORRELATION_KEYS: tuple[str, str, str] = ("run_id", "thread_id", "trace_id")
TRACE_METADATA_KEYS: tuple[str, ...] = TRACE_CORRELATION_KEYS + ("stage", "status")


def _normalize_identifier(value: str | None) -> str:
    if not value:
        return ""
    return value.strip()


def build_langfuse_run_metadata(
    *,
    run_id: str | None = None,
    thread_id: str | None = None,
    trace_id: str | None = None,
    correlation_id: str | None = None,
) -> dict[str, str]:
    normalized_run_id = _normalize_identifier(run_id) or str(uuid.uuid4())
    normalized_thread_id = _normalize_identifier(thread_id) or normalized_run_id
    normalized_trace_id = _normalize_identifier(trace_id) or normalized_run_id
    normalized_correlation_id = _normalize_identifier(correlation_id) or normalized_run_id
    metadata = {
        "run_id": normalized_run_id,
        "thread_id": normalized_thread_id,
        "trace_id": normalized_trace_id,
        "correlation_id": normalized_correlation_id,
    }
    logger.info(
        "Langfuse run metadata prepared run_id=%s thread_id=%s trace_id=%s correlation_id=%s",
        metadata["run_id"],
        metadata["thread_id"],
        metadata["trace_id"],
        metadata["correlation_id"],
    )
    return metadata


def _as_dict(mapping: Mapping[str, Any] | None) -> dict[str, Any]:
    return dict(mapping) if isinstance(mapping, Mapping) else {}


def build_trace_metadata(
    *,
    run_metadata: Any | None = None,
    run_id: str | None = None,
    thread_id: str | None = None,
    trace_id: str | None = None,
    stage: str | None = None,
    status: str | None = None,
    correlation_id: str | None = None,
    metadata: Mapping[str, Any] | None = None,
) -> dict[str, Any]:
    payload = _as_dict(metadata)

    resolved_run_id = _normalize_identifier(run_id or getattr(run_metadata, "run_id", None))
    resolved_thread_id = _normalize_identifier(thread_id or getattr(run_metadata, "thread_id", None))
    resolved_trace_id = _normalize_identifier(trace_id or getattr(run_metadata, "trace_id", None))
    resolved_stage = _normalize_identifier(stage)
    resolved_status = _normalize_identifier(status)
    resolved_correlation_id = _normalize_identifier(correlation_id or getattr(run_metadata, "correlation_id", None))

    if resolved_run_id:
        payload["run_id"] = resolved_run_id
    if resolved_thread_id:
        payload["thread_id"] = resolved_thread_id
    if resolved_trace_id:
        payload["trace_id"] = resolved_trace_id
    if resolved_stage:
        payload["stage"] = resolved_stage
    if resolved_status:
        payload["status"] = resolved_status
    if resolved_correlation_id:
        payload["correlation_id"] = resolved_correlation_id
    return payload


def _call_with_supported_kwargs(target: Any, kwargs: Mapping[str, Any]) -> Any:
    signature = inspect.signature(target)
    accepts_kwargs = any(parameter.kind == inspect.Parameter.VAR_KEYWORD for parameter in signature.parameters.values())
    if accepts_kwargs:
        filtered_kwargs = {key: value for key, value in kwargs.items() if value is not None}
    else:
        accepted = set(signature.parameters.keys())
        filtered_kwargs = {key: value for key, value in kwargs.items() if key in accepted and value is not None}
    return target(**filtered_kwargs)


def _first_callable(target: Any, candidate_names: Iterable[str]) -> Any | None:
    for name in candidate_names:
        candidate = getattr(target, name, None)
        if callable(candidate):
            return candidate
    return None


def _resolve_observation_client(observation: Any) -> Any | None:
    if observation is None:
        return None
    for attr_name in ("langfuse", "client", "_agent_search_langfuse_client"):
        candidate = getattr(observation, attr_name, None)
        if candidate is not None:
            return candidate
    return None


def _load_callback_handler_class() -> Any | None:
    try:
        from langfuse.langchain import CallbackHandler

        return CallbackHandler
    except Exception:  # pragma: no cover - import path can vary by SDK version.
        try:
            from langfuse.callback import CallbackHandler  # type: ignore[attr-defined]

            return CallbackHandler
        except Exception:
            logger.exception("Langfuse tracing enabled but callback handler import failed")
            return None


def _build_langfuse_client(settings: LangfuseSettings) -> Any | None:
    if not settings.enabled:
        logger.info("Langfuse client bootstrap skipped: disabled")
        return None

    if not settings.has_credentials():
        logger.warning(
            "Langfuse client bootstrap skipped: missing credentials public_key_set=%s secret_key_set=%s",
            bool(settings.public_key),
            bool(settings.secret_key),
        )
        return None

    try:
        from langfuse import Langfuse
    except Exception:
        logger.exception("Langfuse tracing enabled but Langfuse module import failed")
        return None

    client_kwargs: dict[str, Any] = {
        "public_key": settings.public_key,
        "secret_key": settings.secret_key,
        "tracing_enabled": True,
        "host": settings.host,
        "environment": settings.environment,
        "release": settings.release,
    }
    signature = inspect.signature(Langfuse.__init__)
    accepted = set(signature.parameters.keys()) - {"self"}
    if "sample_rate" in accepted:
        client_kwargs["sample_rate"] = settings.runtime_sample_rate

    try:
        client = Langfuse(**client_kwargs)
    except Exception:
        logger.exception("Langfuse tracing enabled but Langfuse client init failed")
        return None

    logger.info(
        "Langfuse client initialized host=%s environment=%s release=%s runtime_sample_rate=%s",
        settings.host,
        settings.environment,
        settings.release,
        settings.runtime_sample_rate,
    )
    return client


def get_langfuse_client(*, settings: LangfuseSettings | None = None, force_reinit: bool = False) -> Any | None:
    global _cached_langfuse_client

    resolved_settings = settings or LangfuseSettings.from_env()
    with _cache_lock:
        if not force_reinit and _cached_langfuse_client is not _UNSET:
            return _cached_langfuse_client

        _cached_langfuse_client = _build_langfuse_client(resolved_settings)
        return _cached_langfuse_client


def _should_trace(settings: LangfuseSettings, *, scope: _LangfuseScope, sampling_key: str | None) -> bool:
    if not settings.enabled:
        logger.info("Langfuse tracing disabled enabled=%s", settings.enabled)
        return False

    scope_rate = settings.sample_rate_for_scope(scope)
    sampled = should_sample_rate(scope_rate, sampling_key=sampling_key)
    logger.info(
        "Langfuse trace sampling evaluated scope=%s sample_rate=%s sampled=%s sampling_key_provided=%s",
        scope,
        scope_rate,
        sampled,
        bool(sampling_key),
    )
    return sampled


def build_langfuse_callback_handler(
    *,
    scope: _LangfuseScope = "runtime",
    sampling_key: str | None = None,
    settings: LangfuseSettings | None = None,
) -> Any | None:
    """Create a Langfuse LangChain callback handler when tracing is enabled/sampled."""
    resolved_settings = settings or LangfuseSettings.from_env()
    if not _should_trace(resolved_settings, scope=scope, sampling_key=sampling_key):
        return None

    if not resolved_settings.has_credentials():
        logger.warning(
            "Langfuse tracing enabled but credentials are missing public_key_set=%s secret_key_set=%s",
            bool(resolved_settings.public_key),
            bool(resolved_settings.secret_key),
        )
        return None

    callback_handler_class = _load_callback_handler_class()
    if callback_handler_class is None:
        return None

    langfuse_client = get_langfuse_client(settings=resolved_settings)
    if langfuse_client is None:
        return None

    signature = inspect.signature(callback_handler_class.__init__)
    accepted = set(signature.parameters.keys()) - {"self"}
    kwargs: dict[str, Any] = {}
    if "public_key" in accepted:
        kwargs["public_key"] = resolved_settings.public_key
    if "secret_key" in accepted:
        kwargs["secret_key"] = resolved_settings.secret_key
    if "langfuse_public_key" in accepted:
        kwargs["langfuse_public_key"] = resolved_settings.public_key
    if "langfuse_secret_key" in accepted:
        kwargs["langfuse_secret_key"] = resolved_settings.secret_key
    if "host" in accepted:
        kwargs["host"] = resolved_settings.host
    elif "base_url" in accepted:
        kwargs["base_url"] = resolved_settings.host
    if "environment" in accepted:
        kwargs["environment"] = resolved_settings.environment
    if "release" in accepted:
        kwargs["release"] = resolved_settings.release

    try:
        handler = callback_handler_class(**kwargs)
    except Exception:
        logger.exception("Failed to initialize Langfuse callback handler")
        return None

    setattr(handler, "_agent_search_langfuse_client", langfuse_client)
    logger.info(
        "Langfuse callback handler initialized scope=%s host=%s environment=%s release=%s",
        scope,
        resolved_settings.host,
        resolved_settings.environment,
        resolved_settings.release,
    )
    return handler


def start_langfuse_trace(
    *,
    name: str,
    scope: _LangfuseScope,
    sampling_key: str,
    input_payload: Any | None = None,
    metadata: Mapping[str, Any] | None = None,
    trace_id: str | None = None,
    session_id: str | None = None,
    settings: LangfuseSettings | None = None,
) -> Any | None:
    resolved_settings = settings or LangfuseSettings.from_env()
    if not _should_trace(resolved_settings, scope=scope, sampling_key=sampling_key):
        return None

    langfuse_client = get_langfuse_client(settings=resolved_settings)
    if langfuse_client is None:
        return None

    kwargs: dict[str, Any] = {
        "name": name,
        "input": input_payload,
        "metadata": _as_dict(metadata) or None,
        "id": trace_id,
        "trace_id": trace_id,
        "session_id": session_id,
    }

    trace_builder = _first_callable(langfuse_client, ("trace", "create_trace", "start_trace"))
    if trace_builder is not None:
        try:
            trace = _call_with_supported_kwargs(trace_builder, kwargs)
            setattr(trace, "_agent_search_langfuse_client", langfuse_client)
            logger.info("Langfuse trace started name=%s scope=%s", name, scope)
            return trace
        except Exception:
            logger.exception("Langfuse trace start failed name=%s scope=%s", name, scope)
            return None

    # Langfuse v3 no longer exposes trace constructors on the client.
    # Fall back to a root span/observation with explicit trace context.
    observation_builder = _first_callable(langfuse_client, ("start_span", "start_observation"))
    if observation_builder is None:
        logger.warning("Langfuse trace start skipped; client has no trace/span constructor")
        return None

    observation_kwargs: dict[str, Any] = {
        "name": name,
        "input": input_payload,
        "metadata": _as_dict(metadata) or None,
    }
    trace_context = {k: v for k, v in {"trace_id": trace_id, "session_id": session_id}.items() if v}
    if trace_context:
        observation_kwargs["trace_context"] = trace_context
    try:
        root_observation = _call_with_supported_kwargs(observation_builder, observation_kwargs)
        setattr(root_observation, "_agent_search_langfuse_client", langfuse_client)
        logger.info("Langfuse trace started via span fallback name=%s scope=%s", name, scope)
        return root_observation
    except Exception:
        logger.exception("Langfuse trace/span fallback start failed name=%s scope=%s", name, scope)
        return None


def start_langfuse_span(
    *,
    parent: Any | None,
    name: str,
    input_payload: Any | None = None,
    metadata: Mapping[str, Any] | None = None,
) -> Any | None:
    if parent is None:
        return None

    span_builder = _first_callable(parent, ("span", "create_span", "start_span"))
    if span_builder is None:
        logger.info("Langfuse span start skipped name=%s reason=no_parent_span_builder", name)
        return None

    kwargs: dict[str, Any] = {
        "name": name,
        "input": input_payload,
        "metadata": _as_dict(metadata) or None,
    }
    try:
        span = _call_with_supported_kwargs(span_builder, kwargs)
        logger.info("Langfuse span started name=%s", name)
        return span
    except Exception:
        logger.exception("Langfuse span start failed name=%s", name)
        return None


def end_langfuse_observation(
    observation: Any | None,
    *,
    output_payload: Any | None = None,
    metadata: Mapping[str, Any] | None = None,
) -> None:
    if observation is None:
        return
    update_call = _first_callable(observation, ("update", "set_output"))
    end_call = _first_callable(observation, ("end", "close", "finish"))
    payload = _as_dict(metadata)
    try:
        if update_call is not None:
            update_kwargs = {"output": output_payload, "metadata": payload or None}
            _call_with_supported_kwargs(update_call, update_kwargs)
        # Langfuse v3 span.end() can block while exporting. We already persist output
        # via update(), so avoid blocking end calls on those SDK internals.
        if end_call is not None and getattr(end_call, "__module__", "").startswith("langfuse._client"):
            logger.info("Langfuse observation end skipped for v3 client method=%s", getattr(end_call, "__qualname__", "end"))
            end_call = None
        if end_call is not None:
            end_kwargs = {"output": output_payload, "metadata": payload or None}
            _call_with_supported_kwargs(end_call, end_kwargs)
        logger.info("Langfuse observation closed has_update=%s has_end=%s", update_call is not None, end_call is not None)
    except Exception:
        logger.exception("Langfuse observation close failed")


def record_langfuse_score(
    *,
    parent: Any | None,
    name: str,
    value: float,
    comment: str | None = None,
    metadata: Mapping[str, Any] | None = None,
) -> None:
    if parent is None:
        return
    score_call = _first_callable(parent, ("score", "create_score", "log_score"))
    if score_call is None:
        client = _resolve_observation_client(parent)
        score_call = _first_callable(client, ("score", "create_score", "log_score")) if client is not None else None
    if score_call is None:
        logger.info("Langfuse score skipped name=%s reason=no_score_api", name)
        return

    kwargs: dict[str, Any] = {
        "name": name,
        "value": float(value),
        "score": float(value),
        "comment": comment,
        "metadata": _as_dict(metadata) or None,
    }
    try:
        _call_with_supported_kwargs(score_call, kwargs)
        logger.info("Langfuse score recorded name=%s value=%s", name, value)
    except Exception:
        logger.exception("Langfuse score record failed name=%s", name)


def flush_langfuse_callback_handler(handler: Any | None) -> None:
    """Best-effort flush so traces are quickly available for external harness reads."""
    if handler is None:
        return
    try:
        flush = getattr(handler, "flush", None)
        if callable(flush):
            flush()
            logger.info("Langfuse callback flush completed via handler.flush()")
            return
        client = getattr(handler, "langfuse", None)
        client_flush = getattr(client, "flush", None)
        if callable(client_flush):
            client_flush()
            logger.info("Langfuse callback flush completed via handler.langfuse.flush()")
            return
        attached_client = getattr(handler, "_agent_search_langfuse_client", None)
        attached_client_flush = getattr(attached_client, "flush", None)
        if callable(attached_client_flush):
            attached_client_flush()
            logger.info("Langfuse callback flush completed via attached Langfuse client")
            return
        logger.info("Langfuse callback flush skipped; no flush method found")
    except Exception:
        logger.exception("Langfuse callback flush failed")



from __future__ import annotations

import logging
import os
from dataclasses import dataclass
from typing import Mapping

logger = logging.getLogger(__name__)


def _read_str(env: Mapping[str, str], key: str, default: str) -> str:
    value = env.get(key, "").strip()
    if not value:
        return default
    return value


def _read_bool(env: Mapping[str, str], key: str, default: bool) -> bool:
    raw = env.get(key)
    if raw is None or not raw.strip():
        return default
    normalized = raw.strip().lower()
    if normalized in {"1", "true", "yes", "on"}:
        return True
    if normalized in {"0", "false", "no", "off"}:
        return False
    logger.warning("Invalid boolean config key=%s value=%r default=%s", key, raw, default)
    return default


def _read_float_range(env: Mapping[str, str], key: str, default: float, *, min_value: float, max_value: float) -> float:
    raw = env.get(key)
    if raw is None or not raw.strip():
        return default
    try:
        parsed = float(raw)
    except ValueError:
        logger.warning("Invalid float config key=%s value=%r default=%s", key, raw, default)
        return default
    if parsed < min_value or parsed > max_value:
        logger.warning(
            "Out-of-range float config key=%s value=%s min=%s max=%s default=%s",
            key,
            parsed,
            min_value,
            max_value,
            default,
        )
        return default
    return parsed


@dataclass(frozen=True)
class LangfuseSettings:
    enabled: bool = False
    host: str = "https://cloud.langfuse.com"
    public_key: str = ""
    secret_key: str = ""
    environment: str = "development"
    release: str = "0.1.0"
    runtime_sample_rate: float = 1.0

    @classmethod
    def from_env(cls, env: Mapping[str, str] | None = None) -> LangfuseSettings:
        env_map = os.environ if env is None else env
        settings = cls(
            enabled=_read_bool(env_map, "LANGFUSE_ENABLED", False),
            host=_read_str(
                env_map,
                "LANGFUSE_BASE_URL",
                _read_str(env_map, "LANGFUSE_HOST", "https://cloud.langfuse.com"),
            ),
            public_key=(env_map.get("LANGFUSE_PUBLIC_KEY", "").strip()),
            secret_key=(env_map.get("LANGFUSE_SECRET_KEY", "").strip()),
            environment=_read_str(env_map, "LANGFUSE_ENVIRONMENT", "development"),
            release=_read_str(env_map, "LANGFUSE_RELEASE", "0.1.0"),
            runtime_sample_rate=_read_float_range(
                env_map,
                "LANGFUSE_RUNTIME_SAMPLE_RATE",
                1.0,
                min_value=0.0,
                max_value=1.0,
            ),
        )
        logger.info(
            "Langfuse settings loaded enabled=%s host=%s runtime_sample_rate=%s environment=%s release=%s",
            settings.enabled,
            settings.host,
            settings.runtime_sample_rate,
            settings.environment,
            settings.release,
        )
        return settings

    def has_credentials(self) -> bool:
        return bool(self.public_key and self.secret_key)

    def sample_rate_for_scope(self, scope: str) -> float:
        return self.runtime_sample_rate


def should_sample_rate(sample_rate: float, *, sampling_key: str | None = None) -> bool:
    if sample_rate <= 0.0:
        return False
    if sample_rate >= 1.0:
        return True
    if not sampling_key:
        return True
    digest = hashlib.sha256(sampling_key.encode("utf-8")).hexdigest()
    bucket = int(digest[:8], 16) / 0xFFFFFFFF
    return bucket <= sample_rate
